#!/usr/bin/python
# -*- coding: ascii -*-
"""
Subprocess execution utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import queue
import subprocess as sp
import sys
import time
import typing as tp

from .threading import SDK_TP, SimpleThreadTask


# -->> Definitions <<------------------


# -->> API <<--------------------------
class StreamRedirect:
    """
    Stream redirect utility implementation.
    Allows to store and redirect a given stream.
    """
    def __init__(self, name: str, stream_in: tp.IO[tp.AnyStr], stream_out: tp.IO[tp.AnyStr]) -> None:
        """
        Class initialization.

        :param str name:                Stream redirection name (used for naming threads).
        :param IO[AnyStr] stream_in:    Input stream. Will be stored on buffer and redirected to output.
        :param IO[AnyStr] stream_out:   Output stream.
        """
        self._name  = name
        self._buff  = []
        self._queue = queue.Queue()
        self._src   = stream_in
        self._out   = stream_out
        self._ready = False
        self._start()

    @property
    def buffer(self) -> str:
        """
        Stream buffer.
        """
        return ''.join(self._buff)

    def wait_ready(self) -> None:
        """
        Waits until the stream is finished to continue.
        """
        while not self._ready:
            time.sleep(0.01)

    def _start(self) -> None:
        """
        Initializes the redirection read/write threads.
        """
        SDK_TP.enqueue(SimpleThreadTask(name=f"{self.__class__.__name__}_{self._name}_read", task=self._read))
        SDK_TP.enqueue(SimpleThreadTask(name=f"{self.__class__.__name__}_{self._name}_write", task=self._write))

    def _write(self) -> None:
        """
        Write every line into the output stream and buffer.
        """
        for line in iter(self._queue.get, None):
            self._out.write(line)

    def _read(self) -> None:
        """
        Parses and copies every line of the input stream.
        """
        for line in iter(self._src.readline, b''):
            line = line.decode()
            self._buff.append(line)
            self._queue.put(line)
        self._src.close()
        self._queue.put(None)
        self._ready = True


def execute(cmd: str, pipe: bool = True) -> sp.CompletedProcess:
    """
    Execute the given command as a subprocess.

    :param str cmd:     Command to be executed.
    :param bool pipe:   Pipe output to sys.
    """
    # Prepare
    with sp.Popen(cmd, shell=True, close_fds=True, stdout=sp.PIPE, stderr=sp.PIPE) as proc:
        # Execute
        if pipe:
            # Piping needed...
            # Print header
            print(f"Executing:\n{cmd}\nOutput:")
            # Set piping...
            s_err = StreamRedirect(name="stderr", stream_in=proc.stderr, stream_out=sys.stderr)
            s_out = StreamRedirect(name="stdout", stream_in=proc.stdout, stream_out=sys.stdout)
            # Wait for process and get data
            proc.wait()
            s_err.wait_ready()
            s_out.wait_ready()
            # Get buffers
            err = s_err.buffer
            out = s_out.buffer
        else:
            # Not piping needed...
            out, err = proc.communicate()
            out = '' if (out is None) else out.decode()
            err = '' if (err is None) else err.decode()
        # Return execution result
        return sp.CompletedProcess(args=proc.args, returncode=proc.returncode, stdout=out, stderr=err)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
CRC implementation.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import typing as tp

from .bytes import bitmask, reverse_bits


# -->> Definitions <<------------------


# -->> API <<--------------------------
class CRC:
    """
    Generalized table-driven CRC implementation.

    .. note::
        *   The CRC model definitions can't be changed after initialization.
        *   The CRC configuration values will be adjusted to the defined bit
            size.

    """
    def __init__(self,
                 name: str = "CRC16_CCITT_FALSE", size: int = 16,
                 poly: int = 0x1021, crc_init: int = 0xFFFF, xor_out: int = 0x0000,
                 rev_in: bool = False, rev_out: bool = False
                 ):
        """
        CRC model initialization.

        :param str name:        Model name.
        :param int size:        Size of the CRC in bits.
        :param int poly:        Polynomial generator value.
        :param int crc_init:    Initial value.
        :param int xor_out:     Value to XOR with the CRC result before being returned.
        :param bool rev_in:     If true every input byte will be reversed before being feed to the calculation.
        :param bool rev_out:    If true the final CRC value will be reversed before being returned.
        """
        # Store configurations...
        self._name = name
        self._size = size
        self._rev_in = rev_in
        self._rev_out = rev_out

        # Adjust the values using a bitmask
        self._mask = bitmask(bit=(size - 1), fill=True)
        self._poly = self._mask & poly
        self._xor_out = self._mask & xor_out
        self._crc_init = self._mask & crc_init

        # Generate the lookup table
        self._table = self._compute_lookup_table()

    def __repr__(self) -> str:
        """
        Representation string.
        """
        size = 2 * ((self._size + 7) // 8)
        msg = f'{self.__class__.__name__}(name={self._name}, size={self._size}, ' \
              f'poly=0x{self._poly:0{size:d}X}, ' \
              f'crc_in=0x{self._crc_init:0{size:d}X}, ' \
              f'xor_out=0x{self._xor_out:0{size:d}X}, ' \
              f'ref_in={self._rev_in}, ref_out={self._rev_out})'
        return msg

    # Read only properties
    @property
    def name(self) -> str:
        """
        Model name.
        """
        return self._name

    @property
    def size(self) -> int:
        """
        Model bit size.
        """
        return self._size

    @property
    def poly(self) -> int:
        """
        Model polynomial.
        """
        return self._poly

    @property
    def xor_out(self) -> int:
        """
        Model output XOR value.
        """
        return self._xor_out

    @property
    def crc_init(self) -> int:
        """
        Model initial value.
        """
        return self._crc_init

    @property
    def reverse_in(self) -> bool:
        """
        Model input bit reverse flag.
        """
        return self._rev_in

    @property
    def reverse_out(self) -> bool:
        """
        Model output bit reverse flag.
        """
        return self._rev_out

    @property
    def lookup_table(self) -> tp.List[int]:
        """
        Model pre-computed lookup table.
        """
        return self._table

    def compute(self, data: bytearray, crc_init: int = None) -> int:
        """
        Computes the CRC for the given bytearray.

        :param bytearray data:  Data to compute the CRC over.
        :param int crc_init:    Overrides the default initial CRC value.
            You can use this parameter to perform chained CRC calculations over data blocks.

        :returns: CRC value.
        :rtype: int
        """
        # Select the initial value
        crc_init = self._crc_init if crc_init is None else crc_init

        # Process depending on size
        if self._size >= 8:
            crc = self._compute_normal(data=data, crc_init=crc_init)
        else:
            crc = self._compute_small(data=data, crc_init=crc_init)

        # Prepare the final CRC value and apply XOR
        crc = reverse_bits(value=crc, size=self._size) if self._rev_out else crc
        return crc ^ self._xor_out

    def _compute_lookup_table(self) -> tp.List[int]:
        """
        Generates the lookup table for the current CRC model.

        :returns: Table with 256 pre-computed CRC values.
        :rtype: List[int]
        """
        # Generate the lookup table
        if self._size >= 8:
            return self._lookup_normal()
        return self._lookup_small()

    def _compute_small(self, data: bytearray, crc_init: int) -> int:
        """
        Computes the CRC for model size < 8bit.

        :param bytearray data:  Data to compute the CRC over.
        :param int crc_init:    Overrides the default initial CRC value.

        :returns: CRC value.
        :rtype: int
        """
        # Define required constants
        shift = 8 - self._size
        if self._rev_in:
            # Set the initial value and iterate through bytes
            crc = reverse_bits(value=crc_init, size=self._size)
            for byte in data:
                # Get index and update CRC
                pos = 0xFF & (crc ^ byte)
                crc = self._mask & ((crc >> 8) ^ self._table[pos])
            # Adjust output value
            crc = reverse_bits(value=crc, size=self._size)
        else:
            # Set the initial value and iterate through bytes
            crc = crc_init << shift
            for byte in data:
                # Get index and update CRC
                pos = 0xFF & (crc ^ byte)
                crc = (self._mask << shift) & ((crc << self._size) ^ (self._table[pos] << shift))
            # Adjust output value
            crc >>= shift
        return crc

    def _lookup_small(self) -> tp.List[int]:
        """
        Generates the lookup table for model size < 8bit.

        :returns: Table with 256 pre-computed CRC values.
        :rtype: List[int]
        """
        out = []
        # Define required constants
        shift = (8 - self._size)
        check = bitmask(bit=7)

        # Iterate through all the table values
        for idx in range(256):
            # The current byte is the table index
            byte = reverse_bits(value=idx, size=8) if self._rev_in else idx

            # Compute the CRC value
            for _ in range(8):
                byte = ((byte << 1) ^ (self._poly << shift)) if ((byte & check) != 0) else (byte << 1)

            # Store the value on the table
            byte = (reverse_bits(value=(byte >> shift), size=self._size) << shift) if self._rev_in else byte
            out.append(self._mask & (byte >> shift))
        return out

    def _compute_normal(self, data: bytearray, crc_init: int) -> int:
        """
        Computes the CRC for model size >= 8bit.

        :param bytearray data:  Data to compute the CRC over.
        :param int crc_init:    Overrides the default initial CRC value.

        :returns: CRC value.
        :rtype: int
        """
        # Define required constants
        shift = self._size - 8
        # Set the initial value and iterate through bytes
        crc = crc_init
        for byte in data:
            # Get the byte (reverse if necessary)
            byte = reverse_bits(value=byte, size=8) if self._rev_in else byte
            # Get index and update CRC
            pos = 0xFF & ((crc >> shift) ^ byte)
            crc = self._mask & ((crc << 8) ^ self._table[pos])
        return crc

    def _lookup_normal(self) -> tp.List[int]:
        """
        Generates the lookup table for model size >= 8bit.

        :returns: Table with 256 pre-computed CRC values.
        :rtype: List[int]
        """
        out = []
        # Define required constants
        shift = self._size - 8
        check = bitmask(bit=(self._size - 1))

        # Iterate through all the table values
        for idx in range(256):
            # The current byte is the table index
            byte = idx << shift

            # Compute the CRC value
            for _ in range(8):
                byte = ((byte << 1) ^ self._poly) if ((byte & check) != 0) else (byte << 1)

            # Store the value on the table
            out.append(self._mask & byte)
        return out



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Project build utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import pathlib as pl

from ..utils import execute


# -->> Definitions <<------------------


# -->> API <<--------------------------
def build_cubeide(name: str, config: str, project: pl.Path, workspace: pl.Path) -> None:
    """
    Calls the STM32 CubeIDE headless builder on the specified project.

    ::note:

        - stm32cubeidec executable must be in the PATH.
        - You cannot build if the workspace is already in use.
        - Reference: https://gnu-mcu-eclipse.github.io/advanced/headless-builds/

    :param str name:        Project name.
    :param str config:      Build configuration to be used. Ex: Debug, Release.
    :param Path project:    Path to CubeIDE project.
    :param Path workspace:  Path to CubeIDE workspace.
    """
    cmd = f'stm32cubeidec --launcher.suppressErrors -nosplash ' \
          f'-application org.eclipse.cdt.managedbuilder.core.headlessbuild ' \
          f'-data "{workspace}" -cleanBuild "{name}/{config}" -import "{project}'
    execute(cmd=cmd)


def build_iar(config: str, project: pl.Path) -> None:
    """
    Calls the IAR headless builder on the specified project.

    ::note:

        - IarBuild executable must be in the PATH.

    :param str config:      Build configuration to be used. Ex: Debug, Release.
    :param Path project:    Path to EWARM project.
    """
    cmd = f'IarBuild "{project}" -build "{config}"'
    execute(cmd=cmd)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Timing utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import functools as fc
import time
import typing as tp

from .logger import SDK_LOG


# -->> Definitions <<------------------
#: TyPe definition. Any value.
TPAny      = tp.TypeVar('TPAny')
#: CallBack definition. Any -> Any
CBAny2Any  = tp.Callable[..., TPAny]


# -->> API <<--------------------------
def timer(name: str = None, log: bool = False, precision: int = 6) -> tp.Callable[[CBAny2Any], CBAny2Any]:
    """
    Decorator.
    Used to log the execution time of a function.

    :param str name: Timer name. Used to identify the measurement.
    :param bool log: If true, this info will be captured on the system log. Prints on console otherwise.
    :param float precision: Number of decimals included on elapsed time.

    :return: Decorated function wrapped on timer.
    :rtype: Callable[[Callable[..., RT]], Callable[..., RT]]
    """
    def decorator(func: CBAny2Any) -> CBAny2Any:
        @fc.wraps(func)
        def wrapper(*args, **kwargs) -> TPAny:
            logger = SDK_LOG.info if log else print
            logger(f"{name}: Starting execution...")
            start = time.time()
            ret = func(*args, **kwargs)
            total = elapsed(start)
            logger(f"{name}: Execution time: {total:.{precision}f}[s]")
            return ret
        return wrapper
    return decorator


def elapsed(start: float) -> float:
    """
    Computes the elapsed time since the given timestamp.

    :param float start: Start timestamp.

    :returns: Time elapsed.
    :rtype: float
    """
    return time.time() - start



#!/usr/bin/python
# -*- coding: ascii -*-
"""
SDK threading utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import abc
import functools as fc
import queue
import threading as th
import typing as tp

from .logger import SDK_LOG


# -->> Definitions <<------------------
#: TyPe definition. Any value.
TPAny      = tp.TypeVar('TPAny')
#: CallBack definition. Any -> Any
CBAny2Any  = tp.Callable[..., TPAny]
#: CallBack definition. Any -> None
CBAny2None = tp.Callable[..., None]


# -->> API <<--------------------------
def sync(lock_name: str) -> tp.Callable[[CBAny2Any], CBAny2Any]:
    """
    Decorator.
    Used to wrap a class method with a given lock attribute.

    :param str lock_name: Lock attribute name.

    :returns: Decorated function wrapped on the given lock.
    :rtype: Callable[[Callable[..., RT]], Callable[..., RT]]
    """
    def decorator(func: CBAny2Any) -> CBAny2Any:
        @fc.wraps(func)
        def wrapper(self, *args, **kwargs) -> TPAny:
            lock = self.__getattribute__(lock_name)
            with lock:
                return func(self, *args, **kwargs)
        return wrapper
    return decorator


def get_live_threads(name: str) -> tp.List[th.Thread]:
    """
    Return all the live threads.

    :param str name:  If set, filter all the treads with names that contain the given value.

    :returns: List with named threads.
    :rtype: list
    """
    out = []
    threads = th.enumerate()
    for thread in threads:
        if name in thread.name:
            out.append(thread)
    return out


class AbstractThreadTask(abc.ABC):
    """
    Thread task abstraction.
    Use this class to define how to execute a task inside the ThreadPool.
    """
    @abc.abstractmethod
    def execute(self) -> None:
        """
        Execution called by the ThreadWorkers on the ThreadPool implementation.
        """


class SimpleThreadTask(AbstractThreadTask):
    """
    Simple thread task.
    Accepts a function to be executed by a worker on the ThreadPool.
    """
    def __init__(self, task: CBAny2None, *args, name: str = 'Unnamed', **kwargs) -> None:
        """
        Class initialization.

        :param Callable[..., None] task:    Task functionality.
        :param str name:                    Task name.
        :param args:                        Task arguments.
        :param kwargs:                      Task keyword arguments.
        """
        # Check input
        if not callable(task):
            raise ValueError("The task is not a valid function")

        # Store attributes
        self._name   = name
        self._task   = task
        self._args   = args
        self._kwargs = kwargs

    def __repr__(self) -> str:
        """
        Representation string.
        """
        return f"{self.__class__.__name__}(name={self._name}, task={self._task.__name__})"

    def execute(self) -> None:
        """
        Execution called by the ThreadWorkers on the ThreadPool implementation.
        """
        self._task(*self._args, **self._kwargs)


class ThreadWorker(th.Thread):
    """
    Thread pool worker.
    This represents a single thread on the pool. The thread is set as daemon or not based on
    the pool configurations.
    """
    def __init__(self, name: str, tasks: queue.Queue, timeout: float) -> None:
        """
        Class initialization.

        :param str name:        Worker name.
        :param Queue tasks:     Queue to get the tasks from.
        :param float timeout:   Timeout for waiting for a task.
        """
        super().__init__(name=name)
        self._active    = True
        self._queue     = tasks
        self._timeout   = timeout
        SDK_LOG.debug(f"Worker thread {self.name} created.")

    def stop(self) -> None:
        """
        Stops the thread.
        """
        self._active = False

    def run(self) -> None:
        """
        Runs the worker logic.

        - Blocks until retrieves a task from the tasks queue (locked).
        - Checks that the task is a callable and execute.

        """
        # Worker execution
        SDK_LOG.debug(f"Worker thread {self.name} started.")
        while self._active:
            try:
                task = self._queue.get(timeout=self._timeout)
                if isinstance(task, AbstractThreadTask):
                    try:
                        task.execute()
                    except Exception as ex:
                        SDK_LOG.info(f"Caught exception while running task:\n"
                                     f"> Task   : {task}\n"
                                     f"> Raised : {ex.__class__.__name__} {ex}")
                    finally:
                        self._queue.task_done()
            except queue.Empty:
                pass

        # Worker termination
        SDK_LOG.debug(f"Worker thread {self.name} terminated.")


class ThreadPool:
    """
    Simple thread pool implementation.
    Use queues to coordinate tasks among a set of worker threads.
    """
    def __init__(self, size: int, name: str, timeout: float = 0.1, daemon: bool = True) -> None:
        """
        Class initialization.

        :param int size:        Size of the thread pool.
        :param str name:        Thread pool name. Used as prefix on the workers as <name>_<worker_id>.
        :param float timeout:   Polling time used by the workers while waiting on a get() request
                                on the task queue. This value mainly affects the time used by the
                                threads to terminate when terminate() is called.
        :param bool daemon:     Set to true if the threads should immediately terminate when the
                                main thread exists.
        """
        # Input checking
        if size < 1:
            raise ValueError("Thread pool should have at least one worker")
        if timeout <= 0:
            raise ValueError("Polling timeout should be greater than zero")

        # Set attributes
        self._size      = size
        self._daemon    = daemon
        self._prefix    = name
        self._timeout   = timeout
        self._rlock     = th.RLock()
        self._tasks     = queue.Queue()
        self._workers:  tp.List[ThreadWorker] = []

        # Create workers
        self._create_workers()

    def __del__(self):
        """
        Class destructor.
        Ensures that all the workers are terminated.
        """
        try:
            self.stop()
        except AttributeError:
            pass

    @property
    def size(self) -> int:
        """
        Number of workers.
        """
        return self._size

    @property
    def active(self) -> int:
        """
        Number of active workers.
        """
        return sum([worker.is_alive() for worker in self._workers])

    @sync(lock_name='_rlock')
    def enqueue(self, task: AbstractThreadTask) -> None:
        """
        Enqueue a task on the thread pool to be executed by the workers.

        :param AbstractThreadTask task: Task to added to the queue.
        """
        # Check input
        if not isinstance(task, AbstractThreadTask):
            raise ValueError(f"The task need to be {AbstractThreadTask.__name__} or a subclass of it")
        # Enqueue
        self._tasks.put(task)

    @sync(lock_name='_rlock')
    def stop(self) -> None:
        """
        Wait for the task queue to be completed and stop all the workers.
        """
        SDK_LOG.debug("Waiting for all the tasks to be handled...")
        self._tasks.join()
        SDK_LOG.debug("Terminating worker threads...")
        for worker in self._workers:
            worker.stop()

    def _create_workers(self) -> None:
        """
        Create all the pool workers.
        """
        while len(self._workers) < self._size:
            worker = ThreadWorker(name=f"{self._prefix}_{(len(self._workers) + 1)}", tasks=self._tasks, timeout=self._timeout)
            worker.setDaemon(self._daemon)
            worker.start()
            self._workers.append(worker)


# -->> Instances <<--------------------
#: Embutils internal thread pool
SDK_TP = ThreadPool(size=5, name='EMBUTILS_Thread_')



#!/usr/bin/python
# -*- coding: ascii -*-
"""
COBS encoding/decoding implementation.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import dataclasses as dc
import typing as tp


# -->> Definitions <<------------------


# -->> API <<--------------------------
class COBS:
    """
    Consistent Overhead Byte Stuffing (COBS) encoding/decoding utilities.
    """
    class DecodeException(Exception):
        """
        COBS decoding exception.
        """

    @dc.dataclass
    class Block:
        """
        COBS encoding block.
        """
        code: int
        data: bytearray
        zero: bool = False

    @staticmethod
    def encode(data: bytearray = bytearray()) -> bytearray:
        """
        Encode a byte array using Consistent Overhead Byte Stuffing (COBS).

        .. note::
            * Encoding guarantees non-zero bytes on the output array (besides the terminator).
            * An empty string is encoded to [0x01, 0x00]

        :param bytearray data: Bytes to be encoded.

        :returns: Encoded byte array.
        :rtype: bytearray
        """
        # Prepare
        blocks: tp.List[COBS.Block] = []
        base = data.copy()
        base.append(0x00)

        # Encode
        while base:
            idx = base.find(0x00)
            code = (idx + 1) if (idx != -1) else len(base)
            if code > 0xFE:
                # Extended block:
                blocks.append(COBS.Block(code=0xFF, data=base[0:0xFE]))
                del base[0:0xFE]
            else:
                # Block zero terminated:
                blocks.append(COBS.Block(code=code, data=base[0:(code - 1)]))
                del base[0:code]

        # Process encoding
        out = bytearray()
        for block in blocks:
            out.append(block.code)
            out.extend(block.data)
        out.append(0x00)
        return out

    @staticmethod
    def decode(data: bytearray) -> bytearray:
        """
        Decodes a byte array that was encoded using Consistent Overhead Byte
        Stuffing (COBS).

        :param bytearray data: Bytes to be decoded.

        :returns: Decoded byte array.
        :rtype: bytearray

        :raises COBS.DecodeException: Encoded data is invalid.
        """
        # Prepare
        blocks: tp.List[COBS.Block] = []
        base = data.copy()

        # Decode
        while base:
            # Check if the termination byte is received
            code = base[0]
            if code == 0x00:
                break
            # Add zero to last block, if needed
            if blocks:
                blocks[-1].zero = (blocks[-1].code != 0xFF)
            # Get block and remove bytes
            block = COBS.Block(code=code, data=base[1:code])
            del base[0:code]
            # Check block
            if block.data.find(0x00) != -1:
                raise COBS.DecodeException('Zero byte found in input!')
            if len(block.data) != (code - 1):
                raise COBS.DecodeException('Not enough bytes to process!')
            # Append block
            blocks.append(block)

        # Process decoding
        out = bytearray()
        for block in blocks:
            out.extend(block.data)
            if block.zero:
                out.append(0x00)
        return out



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Serial device implementation classes.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import threading as th
import time
import typing as tp

import serial
import serial.tools.list_ports

from ..utils import EventHook, IntEnum, SimpleThreadTask, sync, elapsed
from ..utils import SDK_LOG, SDK_TP


# -->> Definitions <<------------------


# -->> API <<--------------------------
class Device:
    """
    Serial device implementation wrapper.
    This class includes the USB ID information and allows easily use of looped serial ports for testing.
    """
    #: Default device ID
    DEF_ID = 0xDEADBEEF

    #: Default device settings
    DEF_SETTINGS = {
        'baudrate': 115200,
        'bytesize': 8,
        'stopbits': 1,
        'parity':   'N',
        'timeout':  0.1
        }

    def __init__(self, port: str = None, looped: bool = False, settings: dict = None) -> None:
        """
        Device configuration.
        Applies the serial device settings to the selected port.

        :param str port:        Port name.
        :param bool looped:     Enables the test mode (looped serial).
        :param dict settings:   Serial device configuration.
        """
        # Prepare settings
        if not isinstance(settings, dict):
            settings = self.DEF_SETTINGS

        # Multi-thread safety
        self._lock = th.RLock()

        # Create serial
        self._loop = looped
        if looped:
            self._serial = serial.serial_for_url(url="loop://", do_not_open=True, exclusive=True)
            self._id = self.DEF_ID
        else:
            # Check port
            if port is None:
                raise ValueError("Port is required!")
            # Get ID
            _id = self._id_from_port(port=port)
            if _id is None:
                raise ValueError(f"Port {port} is not connected!")
            # Initialize
            self._serial = serial.Serial()
            self._serial.port = port
            self._id = _id

        # Configure serial
        self._serial.apply_settings(d=settings)
        SDK_LOG.info(f"Serial device created: {self.__repr__()}")

    def __repr__(self) -> str:
        """
        Representation string.
        """
        return f"{self.__class__.__name__}(port={self.port}, id=0x{self.id:08X}, looped={self._loop})"

    def __eq__(self, other: object) -> bool:
        """
        Check if the object is equal to the input.
        """
        if isinstance(other, self.__class__):
            return (self.port == other.port) and (self.id == other.id)
        return False

    def __ne__(self, other: object):
        """
        Check if the object is different to the input.
        """
        return not self.__eq__(other)

    @property
    def port(self) -> str:
        """
        Device port name.
        """
        return self._serial.port

    @property
    def id(self) -> int:
        """
        Device USB ID.
        """
        return self._id

    @property
    def serial(self) -> serial.Serial:
        """
        Serial handler.
        """
        return self._serial

    @property
    def is_open(self) -> bool:
        """
        Returns if the serial device is open.
        """
        return self._serial.is_open

    @sync(lock_name="_lock")
    def open(self) -> bool:
        """
        Tries to open the serial port.

        :returns: True if open, false otherwise.
        :rtype: bool
        """
        try:
            # Check if port is already open
            if self.is_open:
                SDK_LOG.info(f"Port {self.port} already open!")
                return True
            # Try to open it otherwise
            self._serial.open()
            SDK_LOG.info(f"Port {self.port} opened!")
            return True
        except serial.SerialException as ex:
            # Report open issue
            SDK_LOG.error(f"Port {self.port} is unable to connect: {ex}")
            return False

    @sync(lock_name="_lock")
    def close(self) -> None:
        """
        Closes the serial port.
        """
        if self.is_open:
            self.flush()
            self._serial.close()
            SDK_LOG.info(f"Port {self.port} closed!")

    @sync(lock_name="_lock")
    def flush(self) -> None:
        """
        Flushes the serial buffer.
        """
        if self.is_open:
            self._serial.flush()

    @sync(lock_name="_lock")
    def write(self, data: bytearray) -> int:
        """
        Writes the given data through the serial port.

        :param bytearray data: Bytes to be sent through the serial port.

        :returns: Number of bytes sent.
        :rtype: int
        """
        if self.is_open:
            return self._serial.write(data=data)
        return 0

    @sync(lock_name="_lock")
    def read(self, size: int = 1) -> tp.Optional[bytearray]:
        """
        Reads a fixed number of bytes from the serial buffer. The process is
        stopped with error if a timeout is reached before completion.

        :param int size: Number of bytes to read.

        :returns: None if empty or disconnected. Bytearray if bytes received.
        :rtype: Optional[bytearray]
        """
        try:
            if self.is_open:
                return self._serial.read(size=size)
            return None
        except serial.SerialException as ex:
            SDK_LOG.error(f"Port {self.port} presented connection issues while reading: {ex}")
            return None

    @sync(lock_name="_lock")
    def read_until(self, expected: bytes = b'\n', size: int = None) -> tp.Optional[bytearray]:
        """
        Reads bytes from the serial buffer until the expected sequence is found,
        the received bytes exceed the specified limit or a timeout is reached.

        :param bytes expected:  Stop read condition.
        :param int size:        Read array size limit.

        :returns: None if empty or disconnected. Bytearray if bytes received.
        :rtype: Optional[bytearray]
        """
        try:
            if self.is_open:
                return self._serial.read_until(expected=expected, size=size)
            return None
        except serial.SerialException as ex:
            SDK_LOG.error(f"Port {self.port} presented connection issues while reading: {ex}")
            return None

    @staticmethod
    def _id_from_port(port: str) -> tp.Optional[int]:
        """
        Retrieves the USB ID for the given port.

        :param str port: Port name.

        :returns: USB ID
        :rtype: int
        """
        target = None
        devices = serial.tools.list_ports.comports()
        for dev in devices:
            target = dev if (dev.device == port) else target
        return ((target.vid << 16) | target.pid) if target else None


class DeviceList(tp.List[Device]):
    """
    Serial device list implementation.
    This class define mechanisms to scan, compare and filter lists of devices.
    """
    def diff(self, other: 'DeviceList') -> 'DeviceList':
        """
        Get the differences between two serial device lists.

        :param DeviceList other: List to compare with.

        :returns: List containing the differences.
        :rtype: DeviceList
        """
        if self == other:
            return DeviceList()

        # Define base and comparison target
        if len(self) > len(other):
            base = self
            comp = other
        else:
            base = other
            comp = self

        # Perform comparison
        diff = DeviceList()
        for dev in base:
            if dev not in comp:
                diff.append(dev)
        return diff

    def filter(self, port: str = None, dev_id: int = None) -> 'DeviceList':
        """
        Filter serial devices from a given list.

        :param str port:    Port name to be filtered.
        :param int dev_id:  Device ID to be filtered.

        :returns: List containing the filtered devices.
        :rtype: DeviceList
        """
        dev_list = self
        # Filter by port
        if isinstance(port, str):
            dev_list = DeviceList([dev for dev in dev_list if dev.port == port])
        # Filter by ID
        if isinstance(dev_id, int):
            dev_list = DeviceList([dev for dev in dev_list if dev.id == dev_id])
        return dev_list

    @staticmethod
    def scan() -> 'DeviceList':
        """
        Scan the system and return a list with the connected serial devices.

        :returns: List with devices.
        :rtype: DeviceList
        """
        dev_list = DeviceList()

        # Get devices
        dev_scan = serial.tools.list_ports.comports()
        for dev in dev_scan:
            # Not consider items without ID
            if dev.vid is None or dev.pid is None:
                continue
            # Create device
            try:
                new = Device(port=dev.device)
                dev_list.append(new)
            except ValueError as ex:
                SDK_LOG.debug(f"Unable to create a serial device: {ex}")
                continue
        return dev_list


class DeviceScanner:
    """
    Serial device scanner implementation.
    This class define a thread that allows to check periodically for changes on the
    connected serial devices.

    Available events:

    #.  **on_scan_period:** This event is emitted after the scan period is
        completed. Subscribe using callbacks with syntax::

            def <callback>()

    #.  **on_list_change:** This event is emitted when a change is detected
        on the connected device list. Subscribe using callbacks with syntax::

            def <callback>(event: SerialDeviceScanner.Event, devices: SerialDeviceList)

    """
    class Event(IntEnum):
        """
        Serial device scanner event definitions.
        """
        SD_NO_EVENT         = 0x00      # No event
        SD_LIST_CHANGED     = 0x01      # Serial devices list has changed
        SD_PLUGGED_SINGLE   = 0x02      # A single device was plugged
        SD_PLUGGED_MULTI    = 0x03      # Multiple devices were plugged
        SD_REMOVED_SINGLE   = 0x04      # A single device was unplugged
        SD_REMOVED_MULTI    = 0x05      # Multiple devices were unplugged

        @staticmethod
        def get_event(old: DeviceList, new: DeviceList) -> tp.Tuple['DeviceScanner.Event', DeviceList]:
            """
            Compares two serial device lists and return the differences.

            :param DeviceList old: Last scanned device list.
            :param DeviceList new: New scanned device list.

            :returns: Serial device scanner event and device difference list.
            :rtype: Tuple[DeviceScanner.Event, DeviceList]
            """
            # Get difference list and define event
            diff  = old.diff(other=new)
            etype = DeviceScanner.Event
            if not diff:
                event = etype.SD_NO_EVENT
            elif len(new) > len(old):
                event = etype.SD_PLUGGED_MULTI if (len(diff) > 1) else etype.SD_PLUGGED_SINGLE
            elif len(old) > len(new):
                event = etype.SD_REMOVED_MULTI if (len(diff) > 1) else etype.SD_REMOVED_SINGLE
            else:
                event = etype.SD_LIST_CHANGED
            return event, diff

    def __init__(self, period: float = 0.5) -> None:
        """
        Class initialization.

        :param float period:    Scanner refresh period in seconds.
        """
        # Define device lists
        self._dev_list      = DeviceList()    # Devices connected
        self._dev_change    = DeviceList()    # List of devices that triggered the event

        # Define scan period
        self._scan_period = period

        # Define events
        self.on_scan_period = EventHook()
        self.on_list_change = EventHook()

        # Start scanner
        self._active = True
        self._finished  = False
        SDK_TP.enqueue(task=SimpleThreadTask(
            name=f"{self.__class__.__name__}.process",
            task=self._process
            ))

    def __del__(self) -> None:
        """
        Class destructor. Stops the scanner thread.
        """
        if self.is_alive:
            self.stop()

    @property
    def is_alive(self) -> bool:
        """
        Returns if the scan thread is alive.
        """
        return self._active or not self._finished

    @property
    def devices(self) -> DeviceList:
        """
        Connected serial devices list.
        """
        return self._dev_list

    def stop(self) -> None:
        """
        Stops the scanner thread.
        """
        self._active = False
        while self.is_alive:
            time.sleep(0.01)
        SDK_LOG.info('Scanner stopped.')

    def _scan(self) -> None:
        """
        Scanner functionality:

        #. Read the connected devices.
        #. Check for changes on the connected devices.
        #. Raise change event if required, updates current device list.

        """
        # Get the connected devices
        dev_list = DeviceList.scan()

        # If the change callback has items, inform the differences
        if not self.on_list_change.empty:
            event, dev_diff = self.Event.get_event(old=self._dev_change, new=dev_list)
            if event != self.Event.SD_NO_EVENT:
                SDK_TP.enqueue(task=SimpleThreadTask(
                    name=f'{self.__class__.__name__}.on_list_change',
                    task=lambda: self.on_list_change.emit(event=event, dev_diff=dev_diff)
                    ))
                self._dev_change = dev_list

        # Update the connected devices list
        self._dev_list = dev_list
        SDK_TP.enqueue(task=SimpleThreadTask(
            name=f'{self.__class__.__name__}.on_scan_period',
            task=self.on_scan_period.emit
            ))

    def _process(self) -> None:
        """
        Scanner process:

        #. Initialize the scanned devices list.
        #. Ensures the periodic scanner execution.

        """
        SDK_LOG.info('Scanner started.')

        # Initialize connected devices
        self._dev_list = DeviceList.scan()

        # Execute the first scan
        self._scan()

        # Do this periodically
        last_update = time.time()
        while self._active:
            # Wait until the scan period is passed...
            if elapsed(last_update) > self._scan_period:
                last_update = time.time()
                self._scan()

            # Let some time pass
            time.sleep(0.01)

        # Inform finished
        self._finished = True



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Stream implementation.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import abc
import threading as th
import time
import typing as tp

from ..utils import SDK_LOG, SDK_TP, AbstractSerialized, AbstractSerializedCodec, EventHook, SimpleThreadTask, sync
from .device import Device


# -->> Definitions <<------------------


# -->> API <<--------------------------
class AbstractSerializedStreamCodec(AbstractSerializedCodec):
    """
    AbstractSerializedCodec variant for stream usage.
    This class includes the logic required to decode a serialized object from a byte stream.
    """
    @abc.abstractmethod
    def decode_stream(self, device: Device) -> tp.Optional[AbstractSerialized]:
        """
        Decodes a serialized object from a byte stream.

        :param Device device: Stream source.

        :returns: Deserialized object if able, None otherwise.
        :rtype: Optional[AbstractSerialized]

        :raises ConnectionError: Device raised exception while reading.
        """


class Stream:
    """
    This class is used to send and receive serialized objects through a serial device
    in asynchronous way. When a new item is received this class will notify the system
    using events.

    Available events:

    #.  **on_receive:** This event is emitted when an object is received and
        deserialized from the serial device. Subscribe using callbacks with
        syntax::

            def <callback>(item: AbstractSerialized)

    #.  **on_connect:** This event is emitted when the system is able to
        connect to the device. Subscribe using callbacks with syntax::

            def <callback>()

    #.  **on_reconnect:** This event is emitted when the system is able to
        reconnect to the device. Subscribe using callbacks with syntax::

            def <callback>()

    #.  **on_disconnect:** This event is emitted when the system gets
        disconnected from the device. Subscribe using callbacks with syntax::

            def <callback>()

    """
    #: Stream pull period
    PERIOD_PULL_S = 0.005

    #: Stream reconnect try period
    PERIOD_RECONNECT_S = 0.5

    def __init__(self, device: Device, codec: AbstractSerializedStreamCodec) -> None:
        """
        Class initialization.

        :param Device device:                       Serial device handler.
        :param AbstractSerializedStreamCodec codec: Serialized objects codec handler.
        """
        # Initialize
        self._device = device
        self._codec  = codec

        # Events
        self.on_receive     = EventHook()
        self.on_connect     = EventHook()
        self.on_reconnect   = EventHook()
        self.on_disconnect  = EventHook()

        # Debug prints for received
        self.on_receive += lambda item: self._print_debug(item=item, received=True)

        # Multi-thread safety
        self._lock = th.RLock()

        # Start thread
        self._active    = True
        self._finished  = False
        self._paused    = False     # enables pause?
        self._in_pause  = False     # tells if we are actually on pause state
        SDK_TP.enqueue(task=SimpleThreadTask(
            name=f"{self.__class__.__name__}.process",
            task=self._process
            ))

    def __del__(self) -> None:
        """
        Class destructor.
        Stops the stream and the associated serial device.
        """
        if self.is_alive:
            self.stop()
        del self._device

    @property
    def device(self) -> Device:
        """
        Serial device handler.
        """
        return self._device

    @device.setter
    @sync(lock_name="_lock")
    def device(self, device: Device) -> None:
        """
        Serial device handler setter.

        .. note::
            *   This action replaces the current device.
            *   If the new device handler has no port defined
                the current port is preserved.

        :param Device device: Serial device handler.
        """
        self.pause()
        if device.port is not None:
            self._device = device
        self.resume()

    @property
    def codec(self) -> AbstractSerializedStreamCodec:
        """
        Serialized objects codec handler.
        """
        return self._codec

    @property
    def is_alive(self) -> bool:
        """
        Returns if the stream thread is alive.
        """
        return self._active or not self._finished

    @property
    def is_working(self) -> bool:
        """
        Returns if the stream thread is working (not paused).
        """
        return self.is_alive and self._device.is_open and not self._paused

    @sync(lock_name="_lock")
    def send(self, item: AbstractSerialized) -> None:
        """
        Send a serializable item through the serial device.

        :param AbstractSerialized item: Item to send.
        """
        if self.is_working:
            self._print_debug(item=item, received=False)
            self._device.write(data=self._codec.encode(data=item))

    @sync(lock_name="_lock")
    def resume(self) -> None:
        """
        Resume the stream transmission.
        """
        if self.is_alive and self._device.open():
            self._device.flush()
            self._paused = False
            self._in_pause = False
            SDK_LOG.info('Stream resumed.')

    @sync(lock_name="_lock")
    def pause(self) -> None:
        """
        Pauses the stream transmission.
        """
        if self.is_working:
            self._paused = True
            while not self._in_pause:
                time.sleep(0.01)
            self._device.close()
            SDK_LOG.info('Stream paused.')

    @sync(lock_name="_lock")
    def stop(self) -> None:
        """
        Stops the stream transmission and kills the thread.
        """
        self._active = False
        while self.is_alive:
            time.sleep(0.01)
        self._device.close()
        SDK_LOG.info('Stream stopped.')

    def _process(self) -> None:
        """
        Stream process:

        #. Tries to decode items.
        #. If an item is available, emits an event.
        #. Handle disconnection status.

        """
        SDK_LOG.info('Stream started.')

        # Connect to device if not connected
        if not self._device.is_open and self._reconnect():
            SDK_TP.enqueue(task=SimpleThreadTask(
                name=f"{self.__class__.__name__}.on_connect",
                task=self.on_connect.emit
                ))

        # Do this periodically
        while self._active:
            # Handle pause
            if self._paused:
                self._process_paused()
                continue
            # Handle reading
            self._process_active()

        # Inform finished
        self._finished = True

    def _process_paused(self):
        """
        Handle process pause state.
        """
        # Inform that pause started
        if not self._in_pause:
            self._in_pause = True

        # Delay...
        time.sleep(0.01)

    def _process_active(self):
        """
        Handle process active state.
        """
        # Try to receive data, if failed go to reconnection loop
        try:
            item = self._codec.decode_stream(device=self._device)
            if item:
                SDK_TP.enqueue(task=SimpleThreadTask(
                    name=f"{self.__class__.__name__}.on_received",
                    task=lambda: self.on_receive.emit(item=item)
                    ))

        except ConnectionError:
            # Device disconnected...
            SDK_LOG.info(f'Device disconnected: {self._device}')
            SDK_TP.enqueue(task=SimpleThreadTask(
                name=f"{self.__class__.__name__}.on_disconnect",
                task=self.on_disconnect.emit
                ))
            if self._reconnect():
                SDK_TP.enqueue(task=SimpleThreadTask(
                    name=f"{self.__class__.__name__}.on_reconnect",
                    task=self.on_reconnect.emit
                    ))

        # Delay...
        time.sleep(self.PERIOD_PULL_S)

    def _reconnect(self) -> bool:
        """
        Performs a reconnection attempt with the serial device.

        :returns: True if reconnection succeeded, false otherwise.
        :rtype: bool
        """
        status = False
        self._device.close()
        SDK_LOG.info(f'Starting connection attempt on {self._device}')
        while self._active:
            if self._device.open():
                SDK_LOG.info(f'Device {self._device} connected.')
                status = True
                break
            SDK_LOG.info(f'Connection attempt on {self._device} failed.')
            time.sleep(self.PERIOD_RECONNECT_S)
        return status

    @staticmethod
    def _print_debug(item: AbstractSerialized, received: bool) -> None:
        """
        Prints the received/sent items on the SDK logger:

        :param AbstractSerialized item: Item that is being sent/received.
        :param bool received:           Flag to indicate if we are sending/receiving the item.
        """
        action = 'recv' if received else 'sent'
        SDK_LOG.debug(f'Item {action}: {item}')
